/* bootchart-collector
 *
 * Copyright © 2009 Canonical Ltd.
 * Author: Scott James Remnant <scott@netsplit.com>.
 *
 * Modified for Moblin by Auke Kok <auke-jan.h.kok@intel.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/resource.h>
#include <sys/utsname.h>

#include <getopt.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <assert.h>
#include <dirent.h>
#include <limits.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#define VERSION "0.90.1-4 moblin"

/*
 * Typical sizes for a 15-second recording at 0.1s interval = 4MB
 */
#define BUFSIZE 64 * 1024 * 1024


int append_buf (const char *str, size_t len,
		char *outbuf, size_t *outlen);
int copy_buf (int fd, char *outbuf, size_t *outlen);
int flush_buf (int outfd, char *outbuf, size_t *outlen);

int read_file (int fd, const char *uptime, size_t uptimelen,
	       char *outbuf, size_t *outlen);
int read_proc (DIR *proc, const char *uptime, size_t uptimelen,
	       char *outbuf, size_t *outlen);

unsigned long get_uptime (int fd);
void sig_handler (int signum);


int
append_buf (const char *str,
	    size_t      len,
	    char       *outbuf,
	    size_t     *outlen)
{
	assert (len <= BUFSIZE);

	if (*outlen + len > BUFSIZE) {
		perror("No space left in buffer");
		return -1;
	}

	memcpy (outbuf + *outlen, str, len);
	*outlen += len;

	return 0;
}

int
copy_buf (int     fd,
	  char   *outbuf,
	  size_t *outlen)
{
	for (;;) {
		ssize_t len;

		if (*outlen == BUFSIZE) {
			perror("No space left in buffer");
			return -1;
		}

		len = read (fd, outbuf + *outlen, BUFSIZE - *outlen);
		if (len < 0) {
			perror ("read");
			return -1;
		} else if (len == 0)
			break;

		*outlen += len;
	}

	return 0;
}

int
flush_buf (int     outfd,
	   char   *outbuf,
	   size_t *outlen)
{
	size_t writelen = 0;

	while (writelen < *outlen) {
		ssize_t len;

		len = write (outfd, outbuf + writelen, *outlen - writelen);
		if (len < 0) {
			perror ("write");
			exit (1);
		}

		writelen += len;
	}

	*outlen = 0;

	return 0;
}


int
read_file (int         fd,
	   const char *uptime,
	   size_t      uptimelen,
	   char       *outbuf,
	   size_t     *outlen)
{
	lseek (fd, SEEK_SET, 0);

	if (append_buf (uptime, uptimelen, outbuf, outlen) < 0)
		return -1;

	if (copy_buf (fd, outbuf, outlen) < 0)
		return -1;

	if (append_buf ("\n", 1, outbuf, outlen) < 0)
		return -1;

	return 0;
}

int
read_proc (DIR        *proc,
	   const char *uptime,
	   size_t      uptimelen,
	   char       *outbuf,
	   size_t     *outlen)
{
	struct dirent *ent;

	proc = opendir("/proc");
	if (! proc)
		return -1;

	if (append_buf (uptime, uptimelen, outbuf, outlen) < 0)
		return -1;

	while ((ent = readdir (proc)) != NULL) {
		char filename[PATH_MAX];
		int  fd;

		if ((ent->d_name[0] < '0') || (ent->d_name[0] > '9'))
			continue;

		sprintf (filename, "/proc/%s/stat", ent->d_name);

		fd = open (filename, O_RDONLY);
		if (fd < 0)
			continue;

		if (copy_buf (fd, outbuf, outlen) < 0)
			;

		if (close (fd) < 0)
			continue;
	}

	if (append_buf ("\n", 1, outbuf, outlen) < 0)
		return -1;

	if (closedir (proc) < 0)
		return -1;

	return 0;
}


unsigned long
get_uptime (int fd)
{
	char          buf[80];
	ssize_t       len;
	unsigned long u1, u2;

	lseek (fd, SEEK_SET, 0);

	len = read (fd, buf, sizeof buf);
	if (len < 0) {
		perror ("read");
		return 0;
	}

	buf[len] = '\0';

	if (sscanf (buf, "%lu.%lu", &u1, &u2) != 2) {
		perror ("sscanf");
		return 0;
	}

	return u1 * 100 + u2;
}


void
sig_handler (int signum)
{
}

int
main (int   argc,
      char *argv[])
{
	struct sigaction  act;
	sigset_t          mask, oldmask;
	struct rlimit     rlim;
	struct timespec   timeout;
	const char       *output_dir = "/var/log";
	char              filename[PATH_MAX];
	int               sfd, dfd, ufd;
	DIR              *proc;
	FILE             *header;
	int               statfd, diskfd, procfd;
	char             *statbuf, *diskbuf, *procbuf;
	size_t            statlen = 0, disklen = 0, proclen = 0;
	unsigned long     reltime = 0;
	int               rel = 0, len = 300, hz = 20; /* 15 seconds log time */
	char              date[200], cmd[512];
	struct utsname    uts;
	FILE             *file;
	time_t            t;
	int               samples = 0;
	int 		  counter;

	while (1) {
		static struct option opts[] = {
			{ "rel",     0, NULL, 'r' },
			{ "freq",    1, NULL, 'f' },
			{ "samples", 1, NULL, 'n' },
			{ "output",  1, NULL, 'o' },
			{ "help",    0, NULL, 'h' },
			{ 0, 0, NULL, 0 }
		};

		int index = 0, c;

		c = getopt_long(argc, argv, "rf:n:o:t:h", opts, &index);
		if (c == -1)
			break;
		switch (c) {
		case 'r':
			rel = 1;
			break;
		case 'f':
			hz = strtod(optarg, NULL);
			break;
		case 'n':
			len = strtod(optarg, NULL);
			break;
		case 'o':
			output_dir = optarg;
			break;
		case 'h':
			fprintf (stderr, "Usage: %s [OPTIONS]\n", argv[0]);
			fprintf (stderr, " --rel,     -r            Record time relative to recording\n");
			fprintf (stderr, " --freq,    -f N          Sample frequency [%d]\n", hz);
			fprintf (stderr, " --samples, -n N          Stop sampling at [%d] samples\n", len);
			fprintf (stderr, " --output,  -o [PATH]     Path to output files [%s]\n", output_dir);
			fprintf (stderr, " --help,    -h            Display this message\n");
			exit (1);
			break;
		default:
			break;
		}
	}


	if (hz > 1) {
		timeout.tv_sec = 0;
		timeout.tv_nsec = 1000000000 / hz;
	} else {
		timeout.tv_sec = 1;
		timeout.tv_nsec = 0;
	}

	/*
	 * If the kernel executed us through init=/sbin/bootchartd, then
	 * fork:
	 * - parent execs /sbin/init as pid=1
	 * - child logs data
	 */
	if (getpid() == 1) {
		if (fork()) {
			/* parent */
			execl("/sbin/init", "/sbin/init", NULL);
		}
	}

	sigemptyset (&mask);
	sigaddset (&mask, SIGTERM);
	sigaddset (&mask, SIGINT);

	if (sigprocmask (SIG_BLOCK, &mask, &oldmask) < 0) {
		perror ("sigprocmask");
		exit (1);
	}

	act.sa_handler = sig_handler;
	act.sa_flags = 0;
	sigemptyset (&act.sa_mask);

	if (sigaction (SIGTERM, &act, NULL) < 0) {
		perror ("sigaction SIGTERM");
		exit (1);
	}

	if (sigaction (SIGINT, &act, NULL) < 0) {
		perror ("sigaction SIGINT");
		exit (1);
	}

	/* Drop cores if we go wrong */
	if (chdir ("/"))
		;

	rlim.rlim_cur = RLIM_INFINITY;
	rlim.rlim_max = RLIM_INFINITY;

	setrlimit (RLIMIT_CORE, &rlim);

	statbuf = malloc(BUFSIZE);
	if (!statbuf) {
		perror("malloc statbuf");
		exit (1);
	}
	diskbuf = malloc(BUFSIZE);
	if (!diskbuf) {
		perror("malloc diskbuf");
		exit (1);
	}
	procbuf = malloc(BUFSIZE);
	if (!procbuf) {
		perror("malloc procbuf");
		exit (1);
	}

	/*
	 * even though we don't use this DIR *proc just yet, we need to wait
	 * until it is available before we try anything else.
	 */
	counter = 0;
repeat1:
	proc = opendir ("/proc");
	if (! proc) {
		if (counter++ < 100) {
			usleep(100000);
			goto repeat1;
		}
		perror ("opendir /proc");
		exit (1);
	}
	if (closedir (proc) < 0) {
		perror ("close /proc");
		exit (1);
	}


	counter = 0;
repeat2:

	sfd = open ("/proc/stat", O_RDONLY);
	if (sfd < 0) {
		if (counter++ < 100) {
			usleep(100000);
			goto repeat2;
		}
		perror ("open /proc/stat");
		exit (1);
	}

	dfd = open ("/proc/diskstats", O_RDONLY);
	if (dfd < 0) {
		perror ("open /proc/diskstats");
		exit (1);
	}

	ufd = open ("/proc/uptime", O_RDONLY);
	if (ufd < 0) {
		perror ("open /proc/uptime");
		exit (1);
	}



	if (rel) {
		reltime = get_uptime (ufd);
		if (! reltime)
			exit (1);
	}

	for (;;) {
		char          uptime[80];
		size_t        uptimelen;
		unsigned long u;

		u = get_uptime (ufd);
		if (! u)
			exit (1);

		uptimelen = sprintf (uptime, "%lu\n", u - reltime);


		if (read_file (sfd, uptime, uptimelen, statbuf, &statlen) < 0)
			break;

		if (read_file (dfd, uptime, uptimelen, diskbuf, &disklen) < 0)
			break;

		if (read_proc (proc, uptime, uptimelen, procbuf, &proclen) < 0)
			break;

		if ((len > 0) && (samples++ > len))
			break;

		if (pselect (0, NULL, NULL, NULL, &timeout, &oldmask) < 0) {
			if (errno == EINTR) {
				break;
			} else {
				perror ("pselect");
				exit (1);
			}
		}

	}

	sprintf (filename, "%s/proc_stat.log", output_dir);
	counter = 0;
repeat3:
	statfd = open (filename, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (statfd < 0) {
		if (counter++ < 100) {
			usleep(100000);
			goto repeat3;
		}
		perror ("open proc_stat.log");
		exit (1);
	}

	sprintf (filename, "%s/proc_diskstats.log", output_dir);
	diskfd = open (filename, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (diskfd < 0) {
		perror ("open proc_diskstats.log");
		exit (1);
	}

	sprintf (filename, "%s/proc_ps.log", output_dir);
	procfd = open (filename, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (procfd < 0) {
		perror ("open proc_ps.log");
		exit (1);
	}


	if (flush_buf (statfd, statbuf, &statlen) < 0)
		exit (1);
	if (close (statfd) < 0) {
		perror ("close proc_stat.log");
		exit (1);
	}

	if (flush_buf (diskfd, diskbuf, &disklen) < 0)
		exit (1);
	if (close (diskfd) < 0) {
		perror ("close proc_diskstats.log");
		exit (1);
	}

	if (flush_buf (procfd, procbuf, &proclen) < 0)
		exit (1);
	if (close (procfd) < 0) {
		perror ("close proc_ps.log");
		exit (1);
	}

	free(statbuf);
	free(diskbuf);
	free(procbuf);

	if (close (ufd) < 0) {
		perror ("close /proc/uptime");
		exit (1);
	}

	if (close (dfd) < 0) {
		perror ("close /proc/diskstats");
		exit (1);
	}

	if (close (sfd) < 0) {
		perror ("close /proc/stat");
		exit (1);
	}

	/* create header */
	sprintf (filename, "%s/header", output_dir);
	header = fopen (filename, "w");
	if (!header) {
		perror ("open header");
		exit (1);
	}

	if (uname(&uts) < 0) {
		perror("uname");
		exit (1);
	}

	t = time(NULL);
	strftime(date, sizeof(date), "%a, %d %b %Y %H:%M:%S %z", localtime(&t));

	fprintf (header, "version = bootchart %s\n", VERSION);
	fprintf (header, "title = Boot chart for %s (%s)\n", uts.nodename, date);
	fprintf (header, "system.uname = %s %s %s %s\n", uts.sysname, uts.release, uts.version, uts.machine);
	file = fopen("/etc/moblin-release", "r");
	if (!file) {
		sprintf(cmd, "Unknown");
	} else {
		fgets(cmd, sizeof(cmd) - 1, file);
		fclose(file);
	}
	fprintf (header, "system.release = %s", cmd); /* trailing \n in cmd here */
	file = fopen("/proc/cpuinfo", "r");
	if (!file) {
		sprintf(cmd, "Unknown");
	} else {
		char tmp[512];

		sprintf(cmd, "Unknown");
		while (!feof(file)) {
			fgets(tmp, sizeof(tmp) - 1, file);
			if (strstr(tmp, "model name")) {
				strncpy(cmd, tmp + 13, sizeof(tmp) - 14);
				break;
			}
		}
		fclose(file);
	}
	fprintf (header, "system.cpu = %s", cmd); /* trailing \n in cmd here */

	file = fopen("/proc/cmdline", "r");
	if (!file) {
		perror("open /proc/cmdline");
		exit (1);
	}
	fgets(cmd, sizeof(cmd) - 1, file);
	fclose(file);

	fprintf (header, "system.kernel.options = %s", cmd); /* trailing \n in cmd here */

	if (fclose(header) < 0) {
		perror ("fclose header");
		exit (1);
	}

	/* tar up contents */
	sprintf(cmd, "tar cz -C %s -f %s/bootchart.tgz header proc_stat.log proc_diskstats.log proc_ps.log",
		output_dir, output_dir);
	if (system(cmd)) {
		perror(cmd);
		exit (1);
	}

	sprintf (filename, "%s/header", output_dir);
	unlink(filename);
	sprintf (filename, "%s/proc_stat.log", output_dir);
	unlink(filename);
	sprintf (filename, "%s/proc_diskstats.log", output_dir);
	unlink(filename);
	sprintf (filename, "%s/proc_ps.log", output_dir);
	unlink(filename);

	/* generate image file */
	strftime(date, sizeof(date), "%Y%m%d-%H%M", localtime(&t));
	sprintf(cmd, "bootchart -q -f png -o %s/bootchart-%s.png %s", output_dir, date, output_dir);
	if (system(cmd)) {
		perror(cmd);
		exit (1);
	}

	return 0;
}
